from typing import Any

from django import forms
from django.contrib import messages
from django.core.exceptions import ValidationError
from django.db.models import Count
from django.http import HttpRequest
from django.utils.translation import gettext as _

from django_select2.forms import ModelSelect2MultipleWidget, ModelSelect2Widget
from guardian.shortcuts import get_objects_for_user
from material import Fieldset, Layout, Row

from aleksis.core.models import Group, Notification

from .models import AbiGroup, AbiPerson, Comment, ProfileFieldValue, Quote, Ranking, RankingCategory


class PlainTextWidget(forms.Widget):
    def render(self, name, value, attrs=None, renderer=None):
        return value if value is not None else "-"


class AbiPersonForm(forms.ModelForm):
    layout = Layout("name", "person")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["name"].required = False

    class Meta:
        model = AbiPerson
        fields = ["name", "person"]


class AbiGroupForm(forms.ModelForm):
    layout = Layout("name", "members")

    class Meta:
        model = AbiGroup
        fields = ["name", "members"]


class RankingCategoryForm(forms.ModelForm):
    layout = Layout(
        "name", "instruction_text", "group", "number_of_persons", "number_of_persons_per_weight"
    )

    class Meta:
        model = RankingCategory
        fields = [
            "name",
            "instruction_text",
            "group",
            "number_of_persons",
            "number_of_persons_per_weight",
        ]


class ImportForm(forms.Form):
    source_group = forms.ModelChoiceField(queryset=None, label=_("Source group"))
    target_group = forms.ModelChoiceField(queryset=None, label=_("Target group"))

    layout = Layout("source_group", "target_group")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["source_group"].queryset = Group.objects.annotate(
            members_count=Count("members")
        ).filter(members_count__gte=1)
        self.fields["target_group"].queryset = AbiGroup.objects.all()

    def save(self):
        if self.is_valid():
            persons = []
            for member in self.cleaned_data["source_group"].members.all():
                abi_person, __ = AbiPerson.objects.update_or_create(
                    person=member, defaults={"name": f"{member.first_name} {member.last_name}"}
                )
                persons.append(abi_person)

            self.cleaned_data["target_group"].members.add(*persons)


class RankingFormBase(forms.Form):
    def _build_ranking_key(self, category: RankingCategory, weight: int, ranked_by: AbiPerson):
        return f"{category.id}_{weight}_{ranked_by.id}"

    def clean(self) -> dict[str, Any]:
        cleaned_data = super().clean()

        for category, field_names in self.field_layout.items():
            cleaned_data.setdefault(category.field_name, [])
            already_selected_persons = []
            for i, field_name in enumerate(field_names):
                weight = len(field_names) - i
                persons = cleaned_data[field_name] or []
                persons = persons if not isinstance(persons, AbiPerson) else [persons]
                person_key = "_".join(sorted([str(person.id) for person in persons]))

                if persons and person_key in already_selected_persons:
                    raise ValidationError(
                        _(
                            "{category}: You can't select the same"
                            " person(s) multiple times for one category."
                        ).format(category=category.name)
                    )

                if (
                    category.number_of_persons_per_weight > 1
                    and len(persons) > 0
                    and len(persons) != category.number_of_persons_per_weight
                ):
                    raise ValidationError(
                        _(
                            "{category}, {choice}. choice: You have to select "
                            "{number} persons together, not {wrong_number} person(s)."
                        ).format(
                            category=category.name,
                            choice=i + 1,
                            number=category.number_of_persons_per_weight,
                            wrong_number=len(persons),
                        )
                    )

                cleaned_data[category.field_name].append((persons, weight))
                if persons:
                    already_selected_persons.append(person_key)

        return cleaned_data

    def save(self, request: HttpRequest):
        ranked_by = request.user.person.abi_person
        rankings = Ranking.objects.filter(ranked_by=ranked_by)

        if self.is_valid():
            ranking_objects_to_create = []
            persons_to_set = {}
            data = self.cleaned_data

            for category in self.field_layout:
                for persons, weight in data[category.field_name]:
                    existing_ranking = list(
                        filter(
                            lambda ranking: ranking.category == category
                            and ranking.weight == weight,
                            rankings,
                        )
                    )

                    if existing_ranking:
                        if persons:
                            existing_ranking[0].persons.set(persons)
                        else:
                            existing_ranking[0].persons.clear()
                        continue

                    if persons:
                        ranking_key = self._build_ranking_key(category, weight, ranked_by)
                        persons_to_set[ranking_key] = persons

                    ranking_object = Ranking(category=category, weight=weight, ranked_by=ranked_by)

                    ranking_objects_to_create.append(ranking_object)

            new_ranking_objects = Ranking.objects.bulk_create(ranking_objects_to_create)
            for new_ranking_object in new_ranking_objects:
                ranking_key = self._build_ranking_key(
                    new_ranking_object.category,
                    new_ranking_object.weight,
                    new_ranking_object.ranked_by,
                )
                if ranking_key in persons_to_set:
                    new_ranking_object.persons.set(persons_to_set[ranking_key])

            messages.success(request, _("The rankings have been successfully updated."))


def ranking_form_factory(request: HttpRequest) -> forms.Form:
    layout = []
    form_class_attrs = {}
    field_layout = {}

    rankings_by_user = Ranking.objects.filter(ranked_by=request.user.person.abi_person)

    for category in RankingCategory.objects.all():
        field_layout.setdefault(category, [])

        instruction_field = forms.CharField(widget=PlainTextWidget(), required=False)
        instruction_text = category.instruction_text
        if category.number_of_persons_per_weight > 1:
            instruction_text += " " + _(
                _("In this category, you have to select {} persons together.").format(
                    category.number_of_persons_per_weight
                )
            )
        instruction_field.initial = instruction_text
        instruction_field_name = f"{category.field_name}_instruction"
        form_class_attrs[instruction_field_name] = instruction_field

        for i in range(category.number_of_persons):
            if category.number_of_persons_per_weight == 1:
                field = forms.ModelChoiceField(
                    queryset=category.group.members.all(),
                    label=_("{}. choice").format(i + 1),
                    required=False,
                )
                field.widget = ModelSelect2Widget(
                    search_fields=["name__icontains"],
                    attrs={"data-minimum-input-length": 0, "class": "browser-default"},
                )

            else:
                field = forms.ModelMultipleChoiceField(
                    queryset=category.group.members.all(),
                    label=_("{}. choice").format(i + 1),
                    required=False,
                )
                field.widget = ModelSelect2MultipleWidget(
                    search_fields=["name__icontains"],
                    attrs={"data-minimum-input-length": 0, "class": "browser-default"},
                )
            field_name = f"{category.field_name}_{i}"

            ranking_by_user_for_field = rankings_by_user.filter(
                category=category, weight=category.number_of_persons - i
            ).prefetch_related("persons")

            if ranking_by_user_for_field.exists():
                ranking_by_user_for_field = ranking_by_user_for_field.get()
                if category.number_of_persons_per_weight != 1:
                    field.initial = ranking_by_user_for_field.persons.all()
                elif ranking_by_user_for_field.persons.count():
                    field.initial = ranking_by_user_for_field.persons.all()[0]

            form_class_attrs[field_name] = field
            field_layout[category].append(field_name)

        layout.append(
            Fieldset(
                f"{category.name} ({category.group.name})",
                instruction_field_name,
                Row(*field_layout[category]),
            )
        )

    form_class_attrs["layout"] = Layout(*layout)
    form_class_attrs["field_layout"] = field_layout
    form = type(forms.Form)("RankingForm", (RankingFormBase,), form_class_attrs)

    return form


class CommentForm(forms.ModelForm):
    layout = Layout("person", "text")

    class Meta:
        model = Comment
        fields = ["person", "text"]
        widgets = {
            "text": forms.Textarea(attrs={"rows": 4}),
            "person": ModelSelect2Widget(
                search_fields=["name__icontains"],
                attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            ),
        }

    def __init__(self, *args, **kwargs):
        self.person = kwargs.pop("person")
        self.edit = kwargs.pop("edit") if "edit" in kwargs else False
        super().__init__(*args, **kwargs)

        self.allowed_groups = get_objects_for_user(
            self.person.user, "abi.write_comment_for", klass=AbiGroup
        )
        self.allowed_persons = (
            AbiPerson.objects.filter(member_of__in=self.allowed_groups)
            .exclude(pk=self.person.pk)
            .distinct()
        )
        self.allowed_groups_with_review = get_objects_for_user(
            self.person.user, "abi.write_comment_for_with_review", klass=AbiGroup
        )
        self.allowed_persons_with_review = (
            AbiPerson.objects.filter(member_of__in=self.allowed_groups_with_review)
            .exclude(pk=self.person.pk)
            .exclude(person__user=None)
            .distinct()
        )
        allowed_persons = self.allowed_persons | self.allowed_persons_with_review
        self.fields["person"].queryset = allowed_persons

        if self.edit:
            self.fields["person"].disabled = True

    def save(self, commit=True):
        instance = super().save(commit=False)
        instance.author = self.person.abi_person
        if instance.person in self.allowed_persons_with_review:
            instance.needs_review = True
            instance.reviewed = False
            n = Notification(
                sender=_("Abi magazine"),
                recipient=instance.person.person,
                title=_("There are comments to review."),
                description=_(
                    "There are new comments for your profile"
                    " page in the abi magazine we need you to review."
                    " Please decide whether you want to use the comment"
                    " or not by using the menu item 'Abi magazine > Comments'."
                ),
            )
            n.save()
        if commit:
            instance.save()
        return instance


class ProfileFieldValueForm(forms.ModelForm):
    class Meta:
        model = ProfileFieldValue
        fields = ("value",)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["value"].label = self.instance.profile_field.name

    def clean_value(self):
        value = self.cleaned_data["value"]
        if self.instance.profile_field.limit and len(value) > self.instance.profile_field.limit:
            raise ValidationError(
                _("This field has a limit of {} characters.").format(
                    self.instance.profile_field.limit
                )
            )
        return value


class QuoteForm(forms.ModelForm):
    layout = Layout("authors", "text")

    class Meta:
        model = Quote
        fields = ["text", "authors", "submitted_by"]
        widgets = {
            "text": forms.Textarea(attrs={"rows": 4}),
            "authors": ModelSelect2MultipleWidget(
                search_fields=["name__icontains"],
                attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            ),
        }
