import math

from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator
from django.db import models
from django.db.models import Count, Q, QuerySet, Sum
from django.utils.text import slugify
from django.utils.translation import gettext as _

from aleksis.core.mixins import ExtensibleModel
from aleksis.core.models import Person


class AbiPersonQuerySet(QuerySet):
    def annotate_progress(self):
        return self.annotate(
            profile_current=Count(
                "profile__additional_field_values",
                filter=~Q(profile__additional_field_values__value=""),
                distinct=True,
            ),
            rankings_current=Count(
                "done_rankings__category",
                filter=Q(done_rankings__persons__isnull=False),
                distinct=True,
            ),
            number_of_quotes=Count("submitted_quotes__pk", distinct=True),
        )


class AbiPerson(ExtensibleModel):
    objects = AbiPersonQuerySet.as_manager()

    name = models.CharField(
        max_length=255,
        verbose_name=_("Name"),
        help_text=_("Leave this field empty to get the name from the linked person"),
    )
    person = models.OneToOneField(
        to=Person,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        verbose_name=_("AlekSIS person"),
        related_name="abi_person",
    )

    def save(self, *args, **kwargs):
        if not self.name and self.person:
            self.name = f"{self.person.first_name} {self.person.last_name}"
        super().save(*args, **kwargs)

    def __str__(self):
        return self.name

    @property
    def current_progress(self):
        profile_current = (
            (self.profile_current + self.profile.progress_of_own_fields)
            if hasattr(self, "profile")
            else 0
        )
        profile_total = (
            Profile.get_number_of_own_fields() + AdditionalProfileField.objects.all().count()
        )
        profile_percent = profile_current / profile_total * 100
        rankings_current = self.rankings_current
        rankings_total = RankingCategory.objects.all().aggregate(Sum("number_of_persons"))[
            "number_of_persons__sum"
        ]
        rankings_percent = rankings_current / rankings_total * 100 if rankings_total else 0
        return {
            "profile_current": profile_current,
            "profile_total": profile_total,
            "profile_percent": profile_percent,
            "rankings_current": rankings_current,
            "rankings_total": rankings_total,
            "rankings_percent": rankings_percent,
        }

    @property
    def picture_status(self):
        pictures = []
        if hasattr(self, "profile"):
            if self.profile.current_picture:
                pictures.append(_("Current picture"))
            if self.profile.old_picture:
                pictures.append(_("Old picture"))
            if self.profile.additional_picture_1:
                pictures.append(_("Additional picture 1"))
            if self.profile.additional_picture_2:
                pictures.append(_("Additional picture 2"))
        return pictures

    class Meta:
        verbose_name = _("Person (Abi)")
        verbose_name_plural = _("Persons (Abi)")


class AbiGroup(ExtensibleModel):
    name = models.CharField(max_length=255, verbose_name=_("Name"))
    members = models.ManyToManyField(
        to=AbiPerson,
        verbose_name=_("Group members"),
        related_name="member_of",
        null=True,
        blank=True,
    )

    def __str__(self) -> str:
        return self.name

    class Meta:
        verbose_name = _("Group (Abi)")
        verbose_name_plural = _("Groups (Abi)")
        permissions = [
            ("write_comment_for", _("Can write a comment for a person in this group")),
            (
                "write_comment_for_with_review",
                _(
                    "Can write a comment for a person in this group"
                    " which has to be reviewed by the person"
                ),
            ),
        ]


class Profile(ExtensibleModel):
    person = models.OneToOneField(
        to=AbiPerson,
        on_delete=models.CASCADE,
        verbose_name=_("Person"),
        related_name="profile",
        editable=False,
    )
    nickname = models.CharField(max_length=255, blank=True, verbose_name=_("Nickname"))
    current_picture = models.ImageField(
        blank=True, upload_to="abi/current/", verbose_name=_("Current picture")
    )
    old_picture = models.ImageField(
        blank=True, upload_to="abi/old/", verbose_name=_("Picture from the childhood")
    )
    additional_picture_1 = models.ImageField(
        blank=True, upload_to="abi/additional1/", verbose_name=_("Additional picture 1")
    )
    additional_picture_2 = models.ImageField(
        blank=True, upload_to="abi/additional2/", verbose_name=_("Additional picture 2")
    )

    @property
    def progress_of_own_fields(self):
        return len(list(filter(bool, [self.nickname])))

    @classmethod
    def get_number_of_own_fields(cls):
        return 1

    def __str__(self):
        return str(self.person)

    class Meta:
        verbose_name = _("Personal profile")
        verbose_name_plural = _("Personal profiles")
        permissions = (("use_profile", _("Can fill their profile")),)


class AdditionalProfileField(ExtensibleModel):
    name = models.CharField(max_length=255)
    limit = models.PositiveSmallIntegerField(
        default=0, validators=[MaxValueValidator(255)], verbose_name=_("Character limit")
    )
    profile = models.ManyToManyField(
        Profile,
        through="ProfileFieldValue",
        related_name="additional_fields",
        verbose_name=_("Additional Fields"),
    )

    def __str__(self):
        return str(self.name)


class ProfileFieldValue(ExtensibleModel):
    profile = models.ForeignKey(
        Profile,
        on_delete=models.CASCADE,
        related_name="additional_field_values",
        verbose_name=_("Extra Values"),
    )
    profile_field = models.ForeignKey(
        AdditionalProfileField,
        on_delete=models.CASCADE,
        related_name="values",
        verbose_name=_("Fields"),
    )
    value = models.TextField(blank=True)

    @property
    def field_name(self):
        return "profile_" + slugify(self.name).replace("-", "_")

    def __str__(self):
        if self.profile_field is not None:
            return f"{self.profile_field} - {self.profile}"
        else:
            return _("ProfileFieldValue")


class RankingCategory(ExtensibleModel):
    name = models.CharField(max_length=255, verbose_name=_("Name"))
    instruction_text = models.TextField(
        verbose_name=_("Additional instructions for this ranking"),
        blank=True,
        help_text=_("This will be shown under the ranking's name on the ranking page."),
    )
    group = models.ForeignKey(
        to=AbiGroup,
        on_delete=models.CASCADE,
        verbose_name=_("Group"),
        help_text=_("The persons in this group will be selectable for rankings in this category."),
        related_name="ranking_categories",
    )
    number_of_persons = models.PositiveIntegerField(
        default=1, verbose_name=_("Number of selectable persons")
    )
    number_of_persons_per_weight = models.PositiveIntegerField(
        default=1,
        verbose_name=_("Number of selectable persons for one weight (typically one or two)"),
    )

    @property
    def field_name(self):
        return "ranking_" + slugify(self.name).replace("-", "_") + "_" + str(self.group.pk)

    def calculate_ranks(self):
        qs = (
            Ranking.objects.filter(category=self, persons__isnull=False)
            .distinct()
            .prefetch_related("persons")
        )
        persons = {}
        ranks = {}
        for ranking in qs:
            ranks.setdefault(ranking.persons_key, 0)
            persons.setdefault(ranking.persons_key, ranking.persons.all())
            ranks[ranking.persons_key] += ranking.weight

        sorted_ranks = sorted(ranks.items(), key=lambda r: r[1], reverse=True)
        finished_ranks = []
        current_rank = 0
        current_weight = 0
        for rank in sorted_ranks:
            if current_rank == self.number_of_persons:
                break

            if rank[1] != current_weight:
                current_weight = rank[1]
                current_rank += 1
                finished_ranks.append((current_rank, current_weight, persons[rank[0]]))
        print(ranks, sorted_ranks, finished_ranks)
        return finished_ranks

    class Meta:
        verbose_name = _("Ranking category")
        verbose_name_plural = _("Ranking categories")


class Ranking(ExtensibleModel):
    category = models.ForeignKey(
        to=RankingCategory,
        on_delete=models.CASCADE,
        verbose_name=_("Ranking category"),
        related_name="rankings",
    )
    persons = models.ManyToManyField(
        to=AbiPerson,
        blank=True,
        null=True,
        related_name="rankings",
        verbose_name=_("Ranked person"),
    )
    weight = models.PositiveIntegerField(default=1, verbose_name=_("Weight of this ranking"))
    ranked_by = models.ForeignKey(
        to=AbiPerson,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name=_("Ranked by"),
        related_name="done_rankings",
    )

    def clean(self) -> None:
        if self.weight > self.category.number_of_persons:
            raise ValidationError(_("This weight is not available."))

        existing_qs = self.objects.filter(category=self.category, ranked_by=self.person)
        if self.pk:
            existing_qs = existing_qs.exclude(pk=self.pk)
        if existing_qs.count() + 1 > self.category.number_of_persons:
            raise ValidationError(_("You can't rank more persons than allowed."))

        max_weight = math.factorial(self.category.number_of_persons)
        total_weight = existing_qs.aggregate(models.Sum("weight")) + self.weight
        if total_weight > max_weight:
            raise ValidationError(_("You gave more weight than you actually have."))

    @property
    def persons_key(self) -> str:
        return "_".join(sorted([str(p.pk) for p in self.persons.all()]))

    class Meta:
        verbose_name = _("Ranking")
        verbose_name_plural = _("Rankings")
        permissions = (("rank_persons", _("Can rank persons")),)


class Comment(ExtensibleModel):
    person = models.ForeignKey(
        to=AbiPerson,
        on_delete=models.CASCADE,
        verbose_name=_("Recipient"),
        related_name="received_comments",
    )
    author = models.ForeignKey(
        to=AbiPerson,
        on_delete=models.CASCADE,
        verbose_name=_("Author"),
        related_name="written_comments",
    )
    text = models.TextField(
        blank=True,
        verbose_name=_("Comment"),
        help_text=_("This text field supports multiple lines."),
    )
    needs_review = models.BooleanField(default=False, verbose_name=_("Needs review"))
    reviewed = models.BooleanField(default=False, verbose_name=_("Reviewed"))
    accepted = models.BooleanField(default=False, verbose_name=_("Accepted"))

    def __str__(self):
        return self.text

    class Meta:
        verbose_name = _("Comment")
        verbose_name_plural = _("Comments")


class Quote(ExtensibleModel):
    authors = models.ManyToManyField(
        to=AbiPerson,
        verbose_name=_("Author(s)"),
        help_text=_(
            "This can be left empty if the author is unknown or if there is no specific author."
        ),
        blank=True,
        null=True,
        related_name="quotes",
    )
    text = models.TextField(
        verbose_name=_("Quote"),
        help_text=_("You also can include any extra information here like a context if you want."),
    )
    submitted_by = models.ForeignKey(
        to=AbiPerson,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name=_("Submitted by"),
        related_name="submitted_quotes",
    )

    def __str__(self) -> str:
        return self.text

    class Meta:
        verbose_name = _("Quote")
        verbose_name_plural = _("Quotes")
        permissions = (("suggest_quote", _("Can suggest a quote")),)


class ExportedPictures(ExtensibleModel):
    file = models.FileField(upload_to="abi/export/")

    @classmethod
    def get_current(cls):
        return cls.objects.get().file

    @classmethod
    def save_file(cls, file):
        obj = cls.objects.get_or_create()[0]
        obj.file.save("pictures.zip", file)
        return obj

    class Meta:
        verbose_name = _("Exported pictures")
        verbose_name_plural = _("Exported pictures")
