from django.template.loader import render_to_string
from django.utils.translation import gettext as _

import django_tables2 as tables


class AbiPersonTable(tables.Table):
    class Meta:
        attrs = {"class": "highlight"}

    name = tables.LinkColumn("abi_person", args=[tables.A("id")])
    person = tables.Column()
    profile_progress = tables.Column(
        accessor=tables.A("pk"), order_by="profile_current", verbose_name=_("Profile progress")
    )
    rankings_progress = tables.Column(
        accessor=tables.A("pk"), order_by="rankings_current", verbose_name=_("Rankings progress")
    )
    picture_status = tables.Column(
        accessor=tables.A("pk"), order_by="picture_status", verbose_name=_("Picture status")
    )
    number_of_quotes = tables.Column(verbose_name=_("Number of submitted quotes"))
    edit = tables.LinkColumn(
        "edit_abi_person",
        args=[tables.A("id")],
        text=_("Edit"),
        attrs={"a": {"class": "btn-flat waves-effect waves-orange orange-text"}},
        orderable=False,
    )
    delete = tables.LinkColumn(
        "delete_abi_person",
        args=[tables.A("id")],
        text=_("Delete"),
        attrs={"a": {"class": "btn-flat waves-effect waves-red red-text"}},
        orderable=False,
    )

    def render_profile_progress(self, value, record):
        return render_to_string(
            "abi/progress_column.html",
            dict(
                current=record.current_progress["profile_current"],
                total=record.current_progress["profile_total"],
                percent=record.current_progress["profile_percent"],
            ),
        )

    def render_rankings_progress(self, value, record):
        return render_to_string(
            "abi/progress_column.html",
            dict(
                current=record.current_progress["rankings_current"],
                total=record.current_progress["rankings_total"],
                percent=record.current_progress["rankings_percent"],
            ),
        )

    def render_picture_status(self, value, record):
        return render_to_string(
            "abi/picture_status_column.html",
            dict(
                pictures=record.picture_status,
            ),
        )


class AbiGroupTable(tables.Table):
    class Meta:
        attrs = {"class": "highlight"}

    name = tables.LinkColumn("edit_abi_group", args=[tables.A("id")])
    members_count = tables.Column(verbose_name=_("Count of members"))
    edit = tables.LinkColumn(
        "edit_abi_group",
        args=[tables.A("id")],
        text=_("Edit"),
        attrs={"a": {"class": "btn-flat waves-effect waves-orange orange-text"}},
        orderable=False,
    )
    delete = tables.LinkColumn(
        "delete_abi_group",
        args=[tables.A("id")],
        text=_("Delete"),
        attrs={"a": {"class": "btn-flat waves-effect waves-red red-text"}},
        orderable=False,
    )


class RankingCategoryTable(tables.Table):
    class Meta:
        attrs = {"class": "highlight"}

    name = tables.LinkColumn("edit_ranking_category", args=[tables.A("id")])
    group = tables.Column()
    number_of_persons = tables.Column()
    number_of_votes = tables.Column(verbose_name=_("Number of persons which voted"))
    ranking = tables.Column(
        accessor=tables.A("pk"), verbose_name=_("Current ranking"), orderable=False
    )
    edit = tables.LinkColumn(
        "edit_ranking_category",
        args=[tables.A("id")],
        text=_("Edit"),
        attrs={"a": {"class": "btn-flat waves-effect waves-orange orange-text"}},
        orderable=False,
    )
    delete = tables.LinkColumn(
        "delete_ranking_category",
        args=[tables.A("id")],
        text=_("Delete"),
        attrs={"a": {"class": "btn-flat waves-effect waves-red red-text"}},
        orderable=False,
    )

    def render_ranking(self, value, record):
        return render_to_string("abi/ranking_column.html", dict(ranks=record.calculate_ranks()))


class CommentTable(tables.Table):
    class Meta:
        attrs = {"class": "highlight"}

    author = tables.Column()
    person = tables.Column()
    text = tables.Column()
    edit = tables.LinkColumn(
        "edit_comment",
        args=[tables.A("id")],
        text=_("Edit"),
        attrs={"a": {"class": "btn-flat waves-effect waves-orange orange-text"}},
        orderable=False,
    )
    delete = tables.LinkColumn(
        "delete_comment",
        args=[tables.A("id")],
        text=_("Delete"),
        attrs={"a": {"class": "btn-flat waves-effect waves-red red-text"}},
        orderable=False,
    )


class QuotesTable(tables.Table):
    class Meta:
        attrs = {"class": "highlight"}

    text = tables.LinkColumn("edit_quote", args=[tables.A("id")])
    authors = tables.ManyToManyColumn()
    submitted_by = tables.Column()
    edit = tables.LinkColumn(
        "edit_quote",
        args=[tables.A("id")],
        text=_("Edit"),
        attrs={"a": {"class": "btn-flat waves-effect waves-orange orange-text"}},
        orderable=False,
    )
    delete = tables.LinkColumn(
        "delete_quote",
        args=[tables.A("id")],
        text=_("Delete"),
        attrs={"a": {"class": "btn-flat waves-effect waves-red red-text"}},
        orderable=False,
    )


class ReviewCommentTable(CommentTable):
    class Meta:
        exclude = ("person", "edit", "delete")

    review = tables.Column(
        accessor=tables.A("pk"), verbose_name=_("Review status"), orderable=False
    )

    def render_review(self, value, record):
        return render_to_string("abi/comment/review_column.html", dict(comment=record))


class WrittenCommentTable(CommentTable):
    class Meta:
        exclude = ("author",)


class MyQuotesTable(tables.Table):
    class Meta:
        attrs = {"class": "highlight"}

    text = tables.Column()
    authors = tables.ManyToManyColumn()
