from typing import Any

from django.contrib import messages
from django.core.exceptions import PermissionDenied
from django.db.models import Count, OuterRef, Q, Subquery
from django.forms import modelformset_factory
from django.http import Http404, HttpResponse
from django.shortcuts import redirect
from django.urls import reverse, reverse_lazy
from django.utils.decorators import method_decorator
from django.utils.translation import gettext as _
from django.views import View
from django.views.decorators.cache import never_cache
from django.views.generic import FormView, TemplateView
from django.views.generic.detail import DetailView, SingleObjectMixin

from csv_export.views import CSVExportView
from django_tables2 import MultiTableMixin, SingleTableView
from extra_views import ModelFormSetView
from guardian.shortcuts import assign_perm
from material import Layout, Row, Span2, Span3, Span7
from reversion.views import RevisionMixin
from rules.contrib.views import PermissionRequiredMixin

from aleksis.core.mixins import (
    AdvancedCreateView,
    AdvancedDeleteView,
    AdvancedEditView,
    SuccessMessageMixin,
)
from aleksis.core.util.celery_progress import render_progress_page

from .abi_helper import get_date_filename
from .forms import (
    AbiGroupForm,
    AbiPersonForm,
    CommentForm,
    ImportForm,
    ProfileFieldValueForm,
    QuoteForm,
    RankingCategoryForm,
    ranking_form_factory,
)
from .models import (
    AbiGroup,
    AbiPerson,
    AdditionalProfileField,
    Comment,
    ExportedPictures,
    Profile,
    ProfileFieldValue,
    Quote,
    RankingCategory,
)
from .tables import (
    AbiGroupTable,
    AbiPersonTable,
    MyQuotesTable,
    QuotesTable,
    RankingCategoryTable,
    ReviewCommentTable,
    WrittenCommentTable,
)
from .tasks import export_pictures


class AbiPersonListView(PermissionRequiredMixin, SingleTableView):
    """Table of all abi persons."""

    model = AbiPerson
    table_class = AbiPersonTable
    permission_required = "abi.view_abipersons_rule"
    template_name = "abi/abi_person/list.html"

    def get_queryset(self):
        return AbiPerson.objects.annotate_progress()


class AbiPersonDetailView(PermissionRequiredMixin, DetailView):
    """Detail view for abi persons."""

    model = AbiPerson
    permission_required = "abi.view_abiperson_rule"
    template_name = "abi/abi_person/detail.html"

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        if hasattr(self.object, "profile"):
            qs = AdditionalProfileField.objects.all().annotate(
                value=Subquery(
                    ProfileFieldValue.objects.filter(
                        profile=self.object.profile,
                        profile_field=OuterRef("pk"),
                    ).values("value")
                ),
            )
            context["profile_fields"] = qs
            self.object.profile_current = qs.filter(~Q(value="")).count()
            self.object.rankings_current = RankingCategory.objects.filter(
                rankings__persons=self.object
            ).count()
            context["number_of_quotes"] = Quote.objects.filter(submitted_by=self.object).count()
            print(context["profile_fields"])
        else:
            self.object.profile_current = 0
            self.object.rankings_current = 0
        return context


@method_decorator(never_cache, name="dispatch")
class AbiPersonCreateView(PermissionRequiredMixin, AdvancedCreateView):
    """Create view for abi persons."""

    model = AbiPerson
    form_class = AbiPersonForm
    permission_required = "abi.add_abiperson_rule"
    template_name = "abi/abi_person/create.html"
    success_url = reverse_lazy("abi_persons")
    success_message = _("The person has been created.")


@method_decorator(never_cache, name="dispatch")
class AbiPersonEditView(PermissionRequiredMixin, AdvancedEditView):
    """Edit view for abi persons."""

    model = AbiPerson
    form_class = AbiPersonForm
    permission_required = "abi.edit_abiperson_rule"
    template_name = "abi/abi_person/edit.html"
    success_url = reverse_lazy("abi_persons")
    success_message = _("The person has been saved.")


@method_decorator(never_cache, name="dispatch")
class AbiPersonDeleteView(PermissionRequiredMixin, RevisionMixin, AdvancedDeleteView):
    """Delete view for abi persons."""

    model = AbiPerson
    permission_required = "abi.delete_abiperson_rule"
    template_name = "core/pages/delete.html"
    success_url = reverse_lazy("abi_persons")
    success_message = _("The person has been deleted.")


class AbiGroupListView(PermissionRequiredMixin, SingleTableView):
    """Table of all abi groups."""

    model = AbiGroup
    table_class = AbiGroupTable
    permission_required = "abi.view_abigroups_rule"
    template_name = "abi/abi_group/list.html"

    def get_queryset(self):
        return AbiGroup.objects.annotate(members_count=Count("members"))


@method_decorator(never_cache, name="dispatch")
class AbiGroupCreateView(PermissionRequiredMixin, AdvancedCreateView):
    """Create view for abi groups."""

    model = AbiGroup
    form_class = AbiGroupForm
    permission_required = "abi.add_abiaroup_rule"
    template_name = "abi/abi_group/create.html"
    success_url = reverse_lazy("abi_groups")
    success_message = _("The group has been created.")


@method_decorator(never_cache, name="dispatch")
class AbiGroupEditView(PermissionRequiredMixin, AdvancedEditView):
    """Edit view for abi groups."""

    model = AbiGroup
    form_class = AbiGroupForm
    permission_required = "abi.edit_abigroup_rule"
    template_name = "abi/abi_group/edit.html"
    success_url = reverse_lazy("abi_groups")
    success_message = _("The group has been saved.")


@method_decorator(never_cache, name="dispatch")
class AbiGroupDeleteView(PermissionRequiredMixin, RevisionMixin, AdvancedDeleteView):
    """Delete view for abi groups."""

    model = AbiGroup
    permission_required = "abi.delete_abigroup_rule"
    template_name = "core/pages/delete.html"
    success_url = reverse_lazy("abi_groups")
    success_message = _("The group has been deleted.")


@method_decorator(never_cache, name="dispatch")
class AbiGroupImportView(PermissionRequiredMixin, FormView):
    """Import view for abi groups."""

    form_class = ImportForm
    permission_required = "abi.import_abigroup_rule"
    template_name = "abi/abi_group/import.html"
    success_url = reverse_lazy("abi_groups")

    def form_valid(self, form):
        form.save()
        messages.success(self.request, _("The group has been imported."))
        return super().form_valid(form)


class RankingCategoryListView(PermissionRequiredMixin, SingleTableView):
    """Table of all ranking categories."""

    model = RankingCategory
    table_class = RankingCategoryTable
    permission_required = "abi.view_rankingcategories_rule"
    template_name = "abi/ranking/ranking_category/list.html"

    def get_queryset(self):
        return RankingCategory.objects.annotate(
            number_of_votes=Count(
                "rankings__ranked_by", filter=Q(rankings__persons__isnull=False), distinct=True
            )
        )


@method_decorator(never_cache, name="dispatch")
class RankingCategoryCreateView(PermissionRequiredMixin, AdvancedCreateView):
    """Create view for ranking categories."""

    model = RankingCategory
    form_class = RankingCategoryForm
    permission_required = "abi.add_rankingcategory_rule"
    template_name = "abi/ranking/ranking_category/create.html"
    success_url = reverse_lazy("ranking_categories")
    success_message = _("The ranking category has been created.")


@method_decorator(never_cache, name="dispatch")
class RankingCategoryEditView(PermissionRequiredMixin, AdvancedEditView):
    """Edit view for ranking categories."""

    model = RankingCategory
    form_class = RankingCategoryForm
    permission_required = "abi.edit_rankingcategory_rule"
    template_name = "abi/ranking/ranking_category/edit.html"
    success_url = reverse_lazy("ranking_categories")
    success_message = _("The ranking category has been saved.")


@method_decorator(never_cache, name="dispatch")
class RankingCategoryDeleteView(PermissionRequiredMixin, RevisionMixin, AdvancedDeleteView):
    """Delete view for ranking categories."""

    model = RankingCategory
    permission_required = "abi.delete_rankingcategory_rule"
    template_name = "core/pages/delete.html"
    success_url = reverse_lazy("ranking_categories")
    success_message = _("The ranking category has been deleted.")


@method_decorator(never_cache, name="dispatch")
class RankingFormView(PermissionRequiredMixin, FormView):
    """Form view for rankings."""

    permission_required = "abi.rank_persons_rule"
    template_name = "abi/ranking/ranking/form.html"
    success_url = reverse_lazy("ranking_form")

    def get_form_class(self):
        return ranking_form_factory(self.request)

    def form_valid(self, form) -> HttpResponse:
        form.save(self.request)
        return super().form_valid(form)


class CommentListView(PermissionRequiredMixin, MultiTableMixin, SingleTableView):
    """Tables of all comments."""

    model = Comment
    permission_required = "abi.view_comments_rule"
    template_name = "abi/comment/list.html"
    tables = [ReviewCommentTable, WrittenCommentTable]

    def get_tables_data(self):
        qs = super().get_queryset()
        return [
            qs.filter(person=self.request.user.person.abi_person, needs_review=True),
            qs.filter(author=self.request.user.person.abi_person),
        ]


@method_decorator(never_cache, name="dispatch")
class CommentReviewView(PermissionRequiredMixin, SingleObjectMixin, View):
    model = Comment
    permission_required = "abi.review_comment_rule"

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()

        if self.object.reviewed:
            return redirect("comments")

        if kwargs["accept"]:
            self.object.accepted = True
            messages.success(
                request,
                _("Ok, we will use this comment for your profile page in the abi magazine."),
            )
        else:
            self.object.accepted = False
            messages.success(
                request,
                _("Ok, we won't use this comment for your profile page in the abi magazine."),
            )

        self.object.reviewed = True
        self.object.save()

        return redirect("comments")


@method_decorator(never_cache, name="dispatch")
class CommentCreateView(PermissionRequiredMixin, AdvancedCreateView):
    """Create view for comments."""

    model = Comment
    form_class = CommentForm
    permission_required = "abi.add_comment_rule"
    template_name = "abi/comment/create.html"
    success_url = reverse_lazy("comments")
    success_message = _("The comment has been submitted.")

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["person"] = self.request.user.person
        return kwargs

    def form_valid(self, form):
        r = super().form_valid(form)
        assign_perm("abi.view_comment", self.object.author.person.user, self.object)
        assign_perm("abi.change_comment", self.object.author.person.user, self.object)
        assign_perm("abi.delete_comment", self.object.author.person.user, self.object)
        if self.object.needs_review:
            assign_perm("abi.view_comment", self.object.person.person.user, self.object)
        return r


@method_decorator(never_cache, name="dispatch")
class CommentEditView(PermissionRequiredMixin, AdvancedEditView):
    """Edit view for comments."""

    model = Comment
    form_class = CommentForm
    permission_required = "abi.edit_comment_rule"
    template_name = "abi/comment/edit.html"
    success_url = reverse_lazy("comments")
    success_message = _("The comment has been updated.")

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["person"] = self.request.user.person
        kwargs["edit"] = True
        return kwargs


@method_decorator(never_cache, name="dispatch")
class CommentDeleteView(PermissionRequiredMixin, RevisionMixin, AdvancedDeleteView):
    """Delete view for comments."""

    model = Comment
    permission_required = "abi.delete_comment_rule"
    template_name = "core/pages/delete.html"
    success_url = reverse_lazy("comments")
    success_message = _("The comment has been deleted.")


@method_decorator(never_cache, name="dispatch")
class ProfileView(PermissionRequiredMixin, AdvancedEditView):
    model = Profile
    queryset = Profile.objects.all()
    fields = "__all__"
    value_fields_formset = None
    object = None  # noqa
    success_message = _("The profile has been updated successfully!")
    permission_required = "abi.use_profile_rule"

    def get_success_url(self):
        return reverse("update_profile")

    def get_object(self, queryset=None):
        try:
            obj = super().get_object()
        except AttributeError:
            qs = queryset or self.get_queryset()
            try:
                abi_person = AbiPerson.objects.get(person=self.request.user.person)
            except (AttributeError, AbiPerson.DoesNotExist):
                raise PermissionDenied from None
            obj = qs.get_or_create(person=abi_person)[0]
        self.object = obj
        objs = ProfileFieldValue.objects.filter(profile=obj)
        fields = list(objs.values_list("profile_field", flat=True))
        fields_to_create = AdditionalProfileField.objects.all().exclude(pk__in=fields)
        ProfileFieldValue.objects.bulk_create(
            [ProfileFieldValue(profile=obj, profile_field=field) for field in fields_to_create]
        )
        self.value_fields_formset = modelformset_factory(
            model=ProfileFieldValue,
            fields=("value",),
            form=ProfileFieldValueForm,
            extra=0,
        )
        return obj

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if self.request.POST:
            context["additional_fields"] = self.value_fields_formset(
                self.request.POST,
                self.request.FILES,
                queryset=ProfileFieldValue.objects.filter(profile=self.object),
            )
        else:
            context["additional_fields"] = self.value_fields_formset(
                queryset=ProfileFieldValue.objects.filter(profile=self.object)
            )
        return context

    def form_valid(self, form):
        context = self.get_context_data()
        value_fields = context["additional_fields"]

        if value_fields.is_valid():
            value_fields.save()
            return super().form_valid(form)
        else:
            return self.form_invalid(form)


class MyQuotesListView(PermissionRequiredMixin, SingleTableView):
    """Table of all submitted quotes."""

    model = Quote
    table_class = MyQuotesTable
    permission_required = "abi.view_suggested_quotes_rule"
    template_name = "abi/quotes/my.html"

    def get_queryset(self):
        return Quote.objects.filter(submitted_by=self.request.user.person.abi_person)


@method_decorator(never_cache, name="dispatch")
class SuggestQuoteView(PermissionRequiredMixin, AdvancedCreateView):
    """Suggest a quote."""

    model = Quote
    form_class = QuoteForm
    success_message = _("The quote has been sent successfully!")
    permission_required = "abi.suggest_quote"
    template_name = "abi/quotes/suggest.html"

    def get_success_url(self):
        return reverse("my_quotes")

    def form_valid(self, form):
        form.instance.submitted_by = self.request.user.person.abi_person
        r = super().form_valid(form)
        assign_perm("abi.view_quote", self.request.user, self.object)
        return r


@method_decorator(never_cache, name="dispatch")
class CreateQuoteView(PermissionRequiredMixin, AdvancedCreateView):
    """Insertion of quotes."""

    model = Quote
    form_class = QuoteForm
    success_message = _("The quote has been saved successfully!")
    permission_required = "abi.create_quote_rule"
    template_name = "abi/quotes/create.html"

    def get_success_url(self):
        return reverse("quotes")


class QuotesListView(PermissionRequiredMixin, SingleTableView):
    """Table of all added quotes."""

    model = Quote
    table_class = QuotesTable
    permission_required = "abi.view_quotes_rule"
    template_name = "abi/quotes/list.html"


@method_decorator(never_cache, name="dispatch")
class QuoteEditView(PermissionRequiredMixin, AdvancedEditView):
    """Edit view for quotes."""

    model = Quote
    form_class = QuoteForm
    permission_required = "abi.edit_quote_rule"
    template_name = "abi/quotes/edit.html"
    success_url = reverse_lazy("quotes")
    success_message = _("The quote has been saved.")


@method_decorator(never_cache, name="dispatch")
class QuoteDeleteView(PermissionRequiredMixin, RevisionMixin, AdvancedDeleteView):
    """Delete view for quotes."""

    model = Quote
    permission_required = "abi.delete_quote_rule"
    template_name = "core/pages/delete.html"
    success_url = reverse_lazy("quotes")
    success_message = _("The quote has been deleted.")


@method_decorator(never_cache, name="dispatch")
class ConfigureProfileView(PermissionRequiredMixin, SuccessMessageMixin, ModelFormSetView):
    model = AdditionalProfileField
    fields = ("name", "limit")
    factory_kwargs = {"can_delete": True, "can_delete_extra": False}
    success_message = _("The profile questions have been updated successfully.")
    permission_required = "abi.edit_profile_questions_rule"

    def get_object(self, queryset=None):
        self.object = None
        return None

    def get_success_url(self):
        return reverse("configure_profile")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["layout"] = Layout(Row(Span7("name"), Span3("limit"), Span2("DELETE")))
        return context


@method_decorator(never_cache, name="dispatch")
class CommentExportView(CSVExportView):
    model = Comment
    permission_required = "abi.export_comments_rule"
    fields = ("person", "person__person__email", "text", "author", "author__person__email")

    def get_filename(self, queryset):
        return get_date_filename("comments")

    def get_field_value(self, obj, field_name):
        if field_name == "person":
            return obj.person.name
        elif field_name == "author":
            return obj.author.name
        else:
            return super().get_field_value(obj, field_name)

    def get_queryset(self):
        queryset = super().get_queryset()
        return queryset.filter(Q(accepted=True) | Q(needs_review=False))


@method_decorator(never_cache, name="dispatch")
class ProfileExportView(PermissionRequiredMixin, CSVExportView):
    model = Profile
    permission_required = "abi.export_profiles_rule"

    def get_filename(self, queryset):
        return get_date_filename("profiles")

    def get_fields(self, queryset):
        return (
            "person",
            "first_name",
            "last_name",
            "email",
            "nickname",
            "current_picture",
            "old_picture",
        ) + tuple(
            f"field_{val}"
            for val in AdditionalProfileField.objects.all().values_list("id", flat=True)
        )

    def get_queryset(self):
        queryset = super().get_queryset().prefetch_related("additional_field_values")
        for field in AdditionalProfileField.objects.all():
            queryset = queryset.annotate(
                **{
                    f"field_{field.id}": Subquery(
                        ProfileFieldValue.objects.filter(
                            profile=OuterRef("pk"),
                            profile_field=field,
                        ).values("value")
                    ),
                }
            )
        return queryset

    def get_header_name(self, model, field_name):
        if field_name.startswith("field_"):
            id_ = int(field_name.split("_")[-1])
            return AdditionalProfileField.objects.get(id=id_).name
        elif field_name in ("first_name", "last_name", "email"):
            return field_name
        else:
            return super().get_header_name(model, field_name)

    def get_field_value(self, obj, field_name):
        if field_name.startswith("field_"):
            return getattr(obj, field_name)
        elif field_name == "person":
            return obj.person.name
        elif field_name in ("first_name", "last_name", "email"):
            if hasattr(obj.person, "person"):
                return getattr(obj.person.person, field_name, "")
        else:
            return super().get_field_value(obj, field_name)


@method_decorator(never_cache, name="dispatch")
class RankingExportView(CSVExportView):
    model = RankingCategory
    permission_required = "abi.export_rankings_rule"
    fields = (
        "name",
        "number_of_persons",
        "rankings__person",
    )

    def get_filename(self, queryset):
        return get_date_filename("rankings")

    def get_header_name(self, model, field_name):
        if field_name == "rankings__person":
            return _("Winners")
        else:
            return super().get_header_name(model, field_name)

    def get_field_value(self, obj, field_name):
        if field_name == "rankings__person":
            ranks = obj.calculate_ranks()
            rank_items = []
            for rank, __, persons in ranks:
                person_names = " + ".join([p.name for p in persons])
                rank_items.append(f"{person_names} ({rank}.)")
            return ", ".join(rank_items)
        else:
            return super().get_field_value(obj, field_name)


@method_decorator(never_cache, name="dispatch")
class QuoteExportView(CSVExportView):
    model = Quote
    permission_required = "abi.export_quotes_rule"
    fields = (
        "authors",
        "text",
        "submitted_by",
    )

    def get_filename(self, queryset):
        return get_date_filename("quotes")

    def get_field_value(self, obj, field_name):
        if field_name == "submitted_by":
            return getattr(obj.submitted_by, "name", "-")
        else:
            return super().get_field_value(obj, field_name)


@method_decorator(never_cache, name="dispatch")
class ExportOverviewView(PermissionRequiredMixin, TemplateView):
    template_name = "abi/export_overview.html"
    permission_required = "abi.export_rule"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["export_things"] = [
            {
                "title": title,
                "num": num,
                "url": url,
            }
            for title, num, url in [
                (_("Comments"), Comment.objects.all().count(), "export_comment"),
                (_("Profiles"), Profile.objects.all().count(), "export_profile"),
                (_("Quotes"), Quote.objects.all().count(), "export_quote"),
                (_("Rankings"), RankingCategory.objects.all().count(), "export_ranking"),
                (_("Pictures"), Profile.objects.all().count(), "export_picture"),
            ]
        ]

        return context


@method_decorator(never_cache, name="dispatch")
class PictureExportView(PermissionRequiredMixin, View):
    permission_required = "abi.export_profiles_rule"

    def get(self, request, *args, **kwargs):
        result = export_pictures.delay()

        file_url = reverse("get_exported_pictures")

        # Render progress view
        return render_progress_page(
            request,
            result,
            title=_("Progress: Export pictures"),
            back_url=reverse("export_overview"),
            progress_title=_("Build ZIP file …"),
            success_message=_("The ZIP file generation was done successfully."),
            error_message=_("There was a problem while generating the ZIP file."),
            button_url=file_url,
            button_title=_("Download ZIP file"),
            button_icon="download",
        )


@method_decorator(never_cache, name="dispatch")
class GetExportedPicturesView(PermissionRequiredMixin, View):
    permission_required = "abi.export_profiles_rule"

    def get(self, request, *args, **kwargs):
        try:
            zip_file = ExportedPictures.get_current()
        except ExportedPictures.DoesNotExist:
            raise Http404 from None

        return redirect(zip_file.url)
